import json
from datetime import datetime

from .bioc_documents import BiocDocument


class BiocFormatter:
    def build_bioc_format(self, input_vals):
        return {
            "source": "Auto-CORPus (full-text)",
            "date": f'{datetime.today().strftime("%Y%m%d")}',
            "key": "autocorpus_fulltext.key",
            "infons": {},
            "documents": [BiocDocument(input_vals).as_dict()],
        }

    def __init__(self, input_vals):
        self.bioc_output = self.build_bioc_format(input_vals)

    def to_json(self, indentVal=None):
        return json.dumps(self.bioc_output, indent=indentVal, ensure_ascii=False)

    def to_dict(self):
        return self.bioc_output



import argparse
import re
from datetime import datetime
from pathlib import Path

from filetype import is_image
from tqdm import tqdm

from autocorpus.autoCORPus import autoCORPus

parser = argparse.ArgumentParser(prog="PROG")
parser.add_argument(
    "-f", "--filepath", type=str, help="filepath for document/directory to run AC on"
)
parser.add_argument(
    "-t", "--target_dir", type=str, help="target directory"
)  # default autoCORPusOutput
parser.add_argument(
    "-a", "--associated_data", type=str, help="directory of associated data"
)
parser.add_argument(
    "-o",
    "--output_format",
    type=str,
    help=(
        "output format for main text, can be either JSON or XML. "
        "Does not effect tables or abbreviations"
    ),
)
parser.add_argument(
    "-s",
    "--trained_data_set",
    type=str,
    help=(
        "trained dataset to use with pytesseract, must be in the form pytesseract "
        "expects for the lang argument, default eng"
    ),
)

group = parser.add_mutually_exclusive_group()
group.add_argument(
    "-c", "--config", type=str, help="filepath for configuration JSON file"
)
group.add_argument(
    "-d", "--config_dir", type=str, help="directory of configuration JSON files"
)


def get_file_type(file_path: Path) -> str:
    """
    :param file_path: file path to be checked
    :return: "directory", "main_text", "linked_table" or "table_image"
    """
    if file_path.is_dir():
        return "directory"
    elif file_path.suffix == ".html":
        if re.search(r"table_\d+.html", file_path.name):
            return "linked_tables"
        else:
            return "main_text"
    elif is_image(file_path):
        # this should be tidied up to only include the image types which are supported
        # by AC instead of any image files
        return "table_images"
    print(f"unable to identify file type for {file_path}, file will not be processed")
    return ""


def fill_structure(structure, key, ftype, fpath: Path):
    """
    takes the structure dict, if key is not present then creates new entry with default
    vals and adds fpath to correct ftype if key is present then updates the dict with
    the new fpath only

    :param structure: structure dict
    :param key: base file name
    :param ftype: file type (main_text, linked_table, table_image
    :param fpath: full path to the file
    :return: updated structure dct
    """
    if key not in structure:
        structure[key] = {
            "main_text": "",
            "out_dir": "",
            "linked_tables": [],
            "table_images": [],
        }
    if ftype == "main_text" or ftype == "out_dir":
        structure[key][ftype] = str(fpath)
    else:
        structure[key][ftype].append(str(fpath))
    return structure


def read_file_structure(file_path: Path, target_dir: Path):
    """
    takes in any file structure (flat or nested) and groups files, returns a dict of
    files which are all related and the paths to each related file
    :param file_path:
    :return: list of dicts
    """
    if file_path.is_dir():
        structure = {}
        all_fpaths = file_path.rglob("*")
        for fpath in all_fpaths:
            tmp_out = fpath.relative_to(file_path).parent
            out_dir = target_dir / tmp_out
            ftype = get_file_type(fpath)
            base_file = None
            if ftype == "directory":
                continue
            elif ftype == "main_text":
                base_file = re.sub(r"\.html", "", str(fpath))
                structure = fill_structure(structure, base_file, "main_text", fpath)
                structure = fill_structure(structure, base_file, "out_dir", out_dir)
            elif ftype == "linked_tables":
                base_file = re.sub(r"_table_\d+\.html", "", str(fpath))
                structure = fill_structure(structure, base_file, "linked_tables", fpath)
                structure = fill_structure(structure, base_file, "out_dir", out_dir)
            elif ftype == "table_images":
                base_file = re.sub(r"_table_\d+\..*", "", str(fpath))
                structure = fill_structure(structure, base_file, "table_images", fpath)
                structure = fill_structure(structure, base_file, "out_dir", out_dir)
            elif not ftype:
                print(
                    f"cannot determine file type for {fpath}. "
                    "AC will not process this file"
                )
            if base_file in structure:
                structure = fill_structure(structure, base_file, "out_dir", out_dir)
        return structure

    ftype = get_file_type(file_path)
    if ftype == "main_text":
        base_file = re.sub(r"\.html", "", str(file_path)).split("/")[-1]
    if ftype == "linked_tables":
        base_file = re.sub(r"_table_\d+\.html", "", str(file_path)).split("/")[-1]
    if ftype == "table_images":
        base_file = re.sub(r"_table_\d+\..*", "", str(file_path)).split("/")[-1]
    if not ftype:
        raise OSError(
            f"cannot determine file type for {file_path}. AC will not process this file"
        )
    template = {
        base_file: {
            "main_text": "",
            "out_dir": str(target_dir),
            "linked_tables": [],
            "table_images": [],
        }
    }
    template[base_file][get_file_type(file_path)] = str(
        file_path if get_file_type(file_path) == "main_text" else [file_path]
    )
    return template


def main():
    """The main entrypoint for the Auto-CORPus CLI."""
    args = parser.parse_args()
    file_path = Path(args.filepath)
    target_dir = Path(args.target_dir if args.target_dir else "autoCORPus_output")
    config = args.config
    config_dir = args.config_dir  # noqa: F841 ## TODO: Use this variable
    associated_data = args.associated_data  # noqa: F841 ## TODO: Use this variable
    output_format = args.output_format if args.output_format else "JSON"
    trained_data = args.trained_data_set if args.output_format else "eng"

    if not file_path.exists():
        raise FileNotFoundError(f"{file_path} does not exist")
    if not target_dir.exists():
        target_dir.mkdir(parents=True)
    if not target_dir.is_dir():
        raise NotADirectoryError(f"{target_dir} is not a directory")

    structure = read_file_structure(file_path, target_dir)
    pbar = tqdm(structure.keys())
    cdate = datetime.now()

    log_file_path = (
        target_dir / "autoCORPus-log-"
        f"{cdate.day}-{cdate.month}-{cdate.year}-{cdate.hour}-{cdate.minute}"
    )

    with log_file_path.open("w") as log_file:
        log_file.write(
            f"Auto-CORPus log file from {cdate.hour}:{cdate.minute} "
            f"on {cdate.day}/{cdate.month}/{cdate.year}\n"
        )
        log_file.write(f"Input directory provided: {file_path}\n")
        log_file.write(f"Output directory provided: {target_dir}\n")
        log_file.write(f"Config provided: {config}\n")
        log_file.write(f"Output format: {output_format}\n")
        success = []
        errors = []
        for key in pbar:
            pbar.set_postfix(
                {
                    "file": key + "*",
                    "linked_tables": len(structure[key]["linked_tables"]),
                    "table_images": len(structure[key]["table_images"]),
                }
            )
            base_dir = file_path.parent if not file_path.is_dir() else file_path
            try:
                AC = autoCORPus(
                    config,
                    base_dir=str(base_dir),
                    main_text=structure[key]["main_text"],
                    linked_tables=sorted(structure[key]["linked_tables"]),
                    table_images=sorted(structure[key]["table_images"]),
                    trainedData=trained_data,
                )

                out_dir = Path(structure[key]["out_dir"])
                if structure[key]["main_text"]:
                    key = key.replace("\\", "/")
                    if output_format.lower() == "json":
                        with open(
                            out_dir / f"{Path(key).name}_bioc.json",
                            "w",
                            encoding="utf-8",
                        ) as outfp:
                            outfp.write(AC.main_text_to_bioc_json())
                    else:
                        with open(
                            out_dir / f"{Path(key).name}_bioc.xml",
                            "w",
                            encoding="utf-8",
                        ) as outfp:
                            outfp.write(AC.main_text_to_bioc_xml())
                    with open(
                        out_dir / f"{Path(key).name}_abbreviations.json",
                        "w",
                        encoding="utf-8",
                    ) as outfp:
                        outfp.write(AC.abbreviations_to_bioc_json())

                # AC does not support the conversion of tables or abbreviations to XML
                if AC.has_tables:
                    with open(
                        out_dir / f"{Path(key).name}_tables.json", "w", encoding="utf-8"
                    ) as outfp:
                        outfp.write(AC.tables_to_bioc_json())
                success.append(f"{key} was processed successfully.")
            except Exception as e:
                errors.append(f"{key} failed due to {e}.")

        log_file.write(f"{len(success)} files processed.\n")
        log_file.write(f"{len(errors)} files not processed due to errors.\n\n\n")
        log_file.write("\n".join(success) + "\n")
        log_file.write("\n".join(errors) + "\n")
        if errors:
            print(
                "Auto-CORPus has completed processing with some errors. "
                "Please inspect the log file for further details."
            )


if __name__ == "__main__":
    main()



class BioCPassage:
    @classmethod
    def from_title(cls, title, offset):
        title_passage = {
            "section_heading": "",
            "subsection_heading": "",
            "body": title,
            "section_type": [{"iao_name": "document title", "iao_id": "IAO:0000305"}],
        }
        return cls(title_passage, offset)

    def __build_passage(self, passage, offset):
        defaultkeys = ["section_heading", "subsection_heading", "body", "section_type"]
        passage_dict = {
            "offset": offset,
            "infons": {},
            "text": passage["body"],
            "sentences": [],
            "annotations": [],
            "relations": [],
        }
        for key in passage.keys():
            if key not in defaultkeys:
                passage_dict["infons"][key] = passage[key]
        # TODO: currently assumes section_heading and subsection_heading will always exist, should ideally check for existence.
        #  Also doesn't account for subsubsection headings which might exist
        if passage["section_heading"] != "":
            passage_dict["infons"]["section_title_1"] = passage["section_heading"]
        if passage["subsection_heading"] != "":
            passage_dict["infons"]["section_title_2"] = passage["subsection_heading"]
        counter = 1
        for section_type in passage["section_type"]:
            passage_dict["infons"][f"iao_name_{counter}"] = section_type["iao_name"]
            passage_dict["infons"][f"iao_id_{counter}"] = section_type["iao_id"]
            counter += 1

        return passage_dict

    def __init__(self, passage, offset):
        self.offset = 0
        self.passage = self.__build_passage(passage, offset)
        pass

    def as_dict(self):
        return self.passage



import re
from datetime import datetime
from itertools import pairwise, product
from pathlib import Path

from .utils import get_data_element_node, handle_tables, navigate_contents


class table:
    def __table_to_2d(self, t, config):
        """
        transform tables from nested lists to JSON

        Args:
                t: html table, beautiful soup object
                config: configuration dictionary

        Returns:
                table: table in JSON format

        """
        # https://stackoverflow.com/questions/48393253/how-to-parse-table-with-rowspan-and-colspan
        rows = t.find_all("tr")
        # fill colspan and rowspan
        for row in rows:
            for col in row.findAll(["th", "td"]):
                if "colspan" not in col.attrs:
                    col.attrs["colspan"] = 1
                if "rowspan" not in col.attrs:
                    col.attrs["rowspan"] = 1

        # first scan, see how many columns we need
        n_cols = sum(
            [int(i.attrs["colspan"]) for i in t.find("tr").findAll(["th", "td"])]
        )

        # build an empty matrix for all possible cells
        table = [[""] * n_cols for row in rows]

        # fill matrix from row data
        rowspans = {}  # track pending rowspans, column number mapping to count
        for row_idx, row in enumerate(rows):
            span_offset = 0  # how many columns are skipped due to row and colspans
            for col_idx, cell in enumerate(row.findAll(["td", "th"])):
                # adjust for preceding row and colspans
                col_idx += span_offset
                while rowspans.get(col_idx, 0):
                    span_offset += 1
                    col_idx += 1

                # fill table data
                rowspan = rowspans[col_idx] = int(cell.attrs["rowspan"])
                colspan = int(cell.attrs["colspan"])
                # next column is offset by the colspan
                span_offset += colspan - 1
                # value = ''.join(str(x) for x in cell.get_text())
                cCont = cell.contents
                value = ""
                for item in cCont:
                    value += navigate_contents(item)
                # if isinstance(item, bs4.element.NavigableString):
                # 	value += item + " "
                # if isinstance(item, bs4.element.Tag):
                # 	if item.name == "sup" or item.name == "sub":
                # 		value += "<" + item.name + ">"
                # 		value += item.get_text()
                # 		value += "</" + item.name + ">"
                # 	else:
                # 		value += item.get_text()
                # clean the cell
                value = value.strip().replace("\u2009", " ").replace("&#x000a0;", " ")
                value = re.sub(r"\s", " ", value)
                value = re.sub("<\\/?span[^>\n]*>?|<hr\\/>?", "", value)
                value = re.sub("\\n", "", value)
                if value.startswith("(") and value.endswith(")"):
                    value = value[1:-1]
                if re.match(self.pval_regex, value):
                    value = re.sub(
                        r"(\s{0,1})[*××xX](\s{0,1})10(_{0,1})", "e", value
                    ).replace("−", "-")
                if re.match(self.pval_scientific_regex, value):
                    value = re.sub(r"(\s{0,1})[–−-](\s{0,1})", "-", value)
                    value = re.sub(r"(\s{0,1})[eE]", "e", value)
                for drow, dcol in product(range(rowspan), range(colspan)):
                    try:
                        table[row_idx + drow][col_idx + dcol] = value
                        rowspans[col_idx + dcol] = rowspan
                    except IndexError:
                        # rowspan or colspan outside the confines of the table
                        pass
            # update rowspan bookkeeping
            rowspans = {c: s - 1 for c, s in rowspans.items() if s > 1}
        return table

    def __check_superrow(self, row):
        """
        check if the current row is a superrow

        Args:
                row: python list

        Return:
                True/False

        """
        cleaned_row = set(
            i for i in row if (str(i) != "") & (str(i) != "\n") & (str(i) != "None")
        )
        return len(cleaned_row) == 1 and bool(
            re.match("[a-zA-Z]", next(iter(cleaned_row)))
        )

    def __find_format(self, header):
        """
        determine if there exists a splittable pattern in the header cell

        Args:
                header: single header str

        Returns:
                pattern: regex object

        Raises:
                KeyError: Raises an exception.
        """

        if header == "":
            return None
        #     parts = nltk.tokenize.word_tokenize(header)
        a = re.split(r"[:|/,;]", header)
        b = re.findall(r"[:|/,;]", header)
        parts = []
        for i in range(len(b)):
            parts += [a[i], b[i]]
        parts.append(a[-1])

        # identify special character
        special_char_idx = []
        for idx, part in enumerate(parts):
            if part in r":|\/,;":
                special_char_idx.append(idx)

        # generate regex pattern
        if special_char_idx:
            pattern = r""
            for idx in range(len(parts)):
                if idx in special_char_idx:
                    char = parts[idx]
                    pattern += f"({char})"
                else:
                    pattern += r"(\w+)"
            pattern = re.compile(pattern)
            return pattern
        else:
            return None

    def __test_format(self, pattern, s):
        """
        check if the element conforms to the regex pattern

        Args:
                header: single header str
                s: element in string format

        Returns:
                result: bool

        Raises:
                KeyError: Raises an exception.
        """

        if re.search(pattern, s):
            return True
        return False

    def __split_format(self, pattern, s):
        """
        split s according to regex pattern

        Args:
                pattern: regex object
                s: element in string format

        Returns:
                list of substrings

        Raises:
                KeyError: Raises an exception.
        """
        return [i for i in re.split(r"[:|/,;]", s) if i not in r":|\/,;"]

    def __get_headers(self, t, config):
        """
        identify headers from a table

        Args:
                t: BeautifulSoup object of table

        Returns:
                idx_list: a list of header index

        Raises:
                KeyError: Raises an exception.
        """
        idx_list = []
        for idx, row in enumerate(
            get_data_element_node(config["tables"]["data"]["table-row"], t)
        ):
            if get_data_element_node(config["tables"]["data"]["header-element"], row):
                idx_list.append(idx)
            elif "class" in row.attrs:
                if "thead" in row.attrs["class"]:
                    idx_list.append(idx)
        # if no table headers found
        if idx_list == []:
            idx_list = [0]
        return idx_list

    def __get_superrows(self, t):
        """
        determine supperrows in a table

        Args:
                t: BeautifulSoup object of table

        Returns:
                idx_list: a list of superrow index

        """
        idx_list = []
        for idx, row in enumerate(t):
            if idx not in self.__get_headers(t):
                if self.__check_superrow(row):
                    idx_list.append(idx)
        return idx_list

    def __is_number(self, s):
        """
        check if input string is a number

        Args:
                s: input string

        Returns:
                True/False

        """
        try:
            float(s.replace(",", ""))
            return True
        except ValueError:
            return False

    def __is_mix(self, s):
        """
        check if input string is a mix of number and text

        Args:
                s: input string

        Returns:
                True/False

        """
        if any(char.isdigit() for char in s):
            if any(char for char in s if char.isdigit() is False):
                return True
        return False

    def __is_text(self, s):
        """
        check if input string is all text

        Args:
                s: input string

        Returns:
                True/False

        """
        if any(char.isdigit() for char in s):
            return False
        return True

    def __table2json(
        self,
        table_2d,
        header_idx,
        subheader_idx,
        superrow_idx,
        table_num,
        title,
        footer,
        caption,
    ):
        """
        transform tables from nested lists to JSON

        Args:
                table_2d: nested list tables
                header_idx: list of header indices
                subheader_idx: list of subheader indices
                superrow_idx: list of superrow indices
                table_num: table number
                caption: table caption
                footer: table footer

        Returns:
                tables: tables in JSON format

        """
        tables = []
        sections = []
        cur_table = {}
        cur_section = {}

        pre_header = []
        pre_superrow = None
        cur_header = ""
        cur_superrow = ""
        for row_idx, row in enumerate(table_2d):
            if not any([i for i in row if i not in ["", "None"]]):
                continue
            if row_idx in header_idx:
                cur_header = [
                    table_2d[i] for i in next(i for i in subheader_idx if row_idx in i)
                ]
            elif row_idx in superrow_idx:
                cur_superrow = next(i for i in row if i not in ("", "None"))
            else:
                if cur_header != pre_header:
                    sections = []
                    pre_superrow = None
                    cur_table = {
                        "identifier": str(table_num + 1),
                        "title": title,
                        "caption": caption,
                        "columns": cur_header,
                        "section": sections,
                        "footer": footer,
                    }
                    tables.append(cur_table)
                elif cur_header == pre_header:
                    cur_table["section"] = sections
                if cur_superrow != pre_superrow:
                    cur_section = {"section_name": cur_superrow, "results": [row]}
                    sections.append(cur_section)
                elif cur_superrow == pre_superrow:
                    cur_section["results"].append(row)

                pre_header = cur_header
                pre_superrow = cur_superrow

        if len(tables) > 1:
            for table_idx, table in enumerate(tables):
                table["identifier"] += f".{table_idx + 1}"
        return tables

    def __reformat_table_json(self, table_json):
        bioc_format = {
            "source": "Auto-CORPus (tables)",
            "date": f'{datetime.today().strftime("%Y%m%d")}',
            "key": "autocorpus_tables.key",
            "infons": {},
            "documents": [],
        }
        for table in table_json["tables"]:
            if "." in table["identifier"] and self.tableIdentifier:
                tableIdentifier = (
                    self.tableIdentifier + "_" + table["identifier"].split(".")[-1]
                )
            else:
                if self.tableIdentifier:
                    tableIdentifier = self.tableIdentifier
                else:
                    tableIdentifier = table["identifier"].replace(".", "_")
            identifier = tableIdentifier
            offset = 0
            tableDict = {
                "inputfile": self.file_path,
                "id": f"{identifier}",
                "infons": {},
                "passages": [
                    {
                        "offset": 0,
                        "infons": {
                            "section_title_1": "table_title",
                            "iao_name_1": "document title",
                            "iao_id_1": "IAO:0000305",
                        },
                        "text": table["title"],
                    }
                ],
            }
            offset += len(table["title"])
            if "caption" in table.keys() and not table["caption"] == "":
                tableDict["passages"].append(
                    {
                        "offset": offset,
                        "infons": {
                            "section_title_1": "table_caption",
                            "iao_name_1": "caption",
                            "iao_id_1": "IAO:0000304",
                        },
                        "text": ". ".join(table["caption"]),
                    }
                )
                offset += len("".join(table["caption"]))

            if "section" in table.keys():
                rowID = 2
                rsection = []
                this_offset = offset
                for sect in table["section"]:
                    resultsDict = {
                        "table_section_title_1": sect["section_name"],
                        "data_rows": [],
                    }
                    for resultrow in sect["results"]:
                        colID = 1
                        rrow = []
                        for result in resultrow:
                            resultDict = {
                                "cell_id": f"{identifier}.{rowID}.{colID}",
                                "cell_text": result,
                            }
                            colID += 1
                            offset += len(str(result))
                            rrow.append(resultDict)
                        resultsDict["data_rows"].append(rrow)
                        rowID += 1
                    rsection.append(resultsDict)

                columns = []
                for i, column in enumerate(table.get("columns", [])):
                    columns.append(
                        {"cell_id": f"{identifier}.1.{i + 1}", "cell_text": column}
                    )
                tableDict["passages"].append(
                    {
                        "offset": this_offset,
                        "infons": {
                            "section_title_1": "table_content",
                            "iao_name_1": "table",
                            "iao_id_1": "IAO:0000306",
                        },
                        "column_headings": columns,
                        "data_section": rsection,
                    }
                )

            if "footer" in table.keys() and not table["footer"] == "":
                tableDict["passages"].append(
                    {
                        "offset": offset,
                        "infons": {
                            "section_title_1": "table_footer",
                            "iao_name_1": "caption",
                            "iao_id_1": "IAO:0000304",
                        },
                        "text": ". ".join(table["footer"]),
                    }
                )
                offset += len("".join(table["footer"]))
            bioc_format["documents"].append(tableDict)
        return bioc_format

    def __main(self, soup, config):
        soup_tables = handle_tables(config["tables"], soup)

        # remove empty table and other table classes
        pop_list = []
        empty_tables = []

        for i, table in enumerate(soup_tables):
            if "class" in table["node"].attrs:
                if "table-group" in table["node"].attrs["class"]:
                    pop_list.append(i)
            if table["node"].find_all("tbody") == []:
                pop_list.append(i)
                empty_tables.append(table)
        soup_tables = [
            soup_tables[i] for i in range(len(soup_tables)) if i not in pop_list
        ]
        self.empty_tables = []
        for etable in empty_tables:
            if etable["node"].find("table"):
                pass
            # has a table element, not empty
            else:
                etDict = {
                    "title": " ".join(etable["title"]),
                    "caption": " ".join(etable["caption"]),
                    "footer": " ".join(etable["footer"]),
                }
                self.empty_tables.append(etDict)

        # One table
        tables = []
        for table_num, table in enumerate(soup_tables):
            # remove empty table header
            if table["node"].find("td", "thead-hr"):
                table["node"].find("td", "thead-hr").parent.extract()

            header_idx = self.__get_headers(table["node"], config)

            # span table to single-cells
            table_2d = self.__table_to_2d(table["node"], config)

            # find superrows
            superrow_idx = []
            if table_2d is not None:
                for row_idx, row in enumerate(table_2d):
                    if row_idx not in header_idx:
                        if self.__check_superrow(row):
                            superrow_idx.append(row_idx)

            # identify section names in index column
            if superrow_idx == []:
                first_col = [row[0] for row in table_2d]
                first_col_vals = [
                    i for i in first_col if first_col.index(i) not in header_idx
                ]
                unique_vals = set([i for i in first_col_vals if i not in ["", "None"]])
                if len(unique_vals) <= len(first_col_vals) / 2:
                    section_names = list(unique_vals)
                    for i in section_names:
                        superrow_idx.append(first_col.index(i))
                    n_cols = len(table_2d[0])
                    for idx, val in zip(superrow_idx, section_names):
                        table_2d = table_2d[:idx] + [[val] * n_cols] + table_2d[idx:]
                    # update superrow_idx after superrow insertion
                    superrow_idx = []
                    first_col = [row[0] for row in table_2d]
                    for i in section_names:
                        superrow_idx.append(first_col.index(i))
                    for row in table_2d:
                        row.pop(0)

            # Identify subheaders
            value_idx = [
                i for i in range(len(table_2d)) if i not in header_idx + superrow_idx
            ]
            col_type = []
            for col_idx in range(len(table_2d[0])):
                cur_col = [i[col_idx] for i in table_2d]
                num_cnt = 0
                txt_cnt = 0
                mix_cnt = 0
                for cell in cur_col:
                    cell = str(cell).lower()
                    if cell in [
                        "none",
                        "",
                        "-",
                    ]:
                        continue
                    elif self.__is_number(cell):
                        num_cnt += 1
                    elif self.__is_mix(cell):
                        mix_cnt += 1
                    elif self.__is_text(cell):
                        txt_cnt += 1
                if max(num_cnt, txt_cnt, mix_cnt) == num_cnt:
                    col_type.append("num")
                elif max(num_cnt, txt_cnt, mix_cnt) == txt_cnt:
                    col_type.append("txt")
                else:
                    col_type.append("mix")
            subheader_idx = []
            for row_idx in value_idx:
                cur_row = table_2d[row_idx]
                unmatch_cnt = 0
                for col_idx in range(len(cur_row)):
                    cell = str(cur_row[col_idx]).lower()
                    if (
                        self.__is_text(cell)
                        and col_type[col_idx] != "txt"
                        and cell
                        not in [
                            "none",
                            "",
                            "-",
                        ]
                    ):
                        unmatch_cnt += 1
                if unmatch_cnt >= len(cur_row) / 2:
                    subheader_idx.append(row_idx)
            header_idx += subheader_idx

            subheader_idx = []
            tmp = [header_idx[0]]
            for i, j in pairwise(header_idx):
                if j == i + 1:
                    tmp.append(j)
                else:
                    subheader_idx.append(tmp)
                    tmp = [j]
            subheader_idx.append(tmp)

            # convert to float
            for row in table_2d:
                for cell in range(len(row)):
                    try:
                        row[cell] = float(
                            row[cell]
                            .replace("−", "-")
                            .replace("–", "-")
                            .replace(",", "")
                        )
                    except Exception:
                        row[cell] = row[cell]

            cur_table = self.__table2json(
                table_2d,
                header_idx,
                subheader_idx,
                superrow_idx,
                table_num,
                table["title"],
                table["footer"],
                table["caption"],
            )
            # merge headers
            sep = "|"
            for table in cur_table:
                headers = table["columns"]
                new_header = []
                if not headers:
                    continue
                for col_idx in range(len(headers[0])):
                    new_element = ""
                    for r_idx in range(len(headers)):
                        new_element += str(headers[r_idx][col_idx]) + sep
                    new_element = new_element.rstrip(sep)
                    new_header.append(new_element)
                table["columns"] = new_header

            tables += cur_table

        table_json = {"tables": tables}
        table_json = self.__reformat_table_json(table_json)
        return table_json

    def __init__(self, soup, config, file_name, base_dir):
        self.file_path = file_name
        file_name = Path(file_name).name
        self.tableIdentifier = None
        self.base_dir = base_dir
        if re.search(r"_table_\d+\.html", file_name):
            self.tableIdentifier = file_name.split("/")[-1].split("_")[-1].split(".")[0]
        self.pval_regex = (
            r"((\d+\.\d+)|(\d+))(\s?)[*××xX](\s{0,1})10[_]{0,1}([–−-])(\d+)"
        )
        self.pval_scientific_regex = (
            r"((\d+.\d+)|(\d+))(\s{0,1})[eE](\s{0,1})([–−-])(\s{0,1})(\d+)"
        )
        self.tables = self.__main(soup, config)
        pass

    def to_dict(self):
        return self.tables, self.empty_tables



from pathlib import Path

from .bioc_passages import BioCPassage


class BiocDocument:
    def build_passages(self, dataStore):
        seen_headings = []
        passages = [BioCPassage.from_title(dataStore.main_text["title"], 0).as_dict()]
        if dataStore.main_text["title"] not in seen_headings:
            offset = len(dataStore.main_text["title"])
            seen_headings.append(dataStore.main_text["title"])
        for passage in dataStore.main_text["paragraphs"]:
            passage_obj = BioCPassage(passage, offset)
            passages.append(passage_obj.as_dict())
            offset += len(passage["body"])
            if passage["subsection_heading"] not in seen_headings:
                offset += len(passage["subsection_heading"])
                seen_headings.append(passage["subsection_heading"])
            if passage["section_heading"] not in seen_headings:
                offset += len(passage["section_heading"])
                seen_headings.append(passage["section_heading"])
        return passages

    def build_template(self, dataStore):
        return {
            "id": Path(dataStore.file_path).name.split(".")[0],
            "inputfile": dataStore.file_path,
            "infons": {},
            "passages": self.build_passages(dataStore),
            "annotations": [],
            "relations": [],
        }

    def __init__(self, input):
        self.document = self.build_template(input)
        pass

    def as_dict(self):
        return self.document



import argparse
import json
import sys
from pathlib import Path

from bioc import biocjson, biocxml
from bs4 import BeautifulSoup

from .abbreviation import abbreviations
from .bioc_formatter import BiocFormatter
from .section import section
from .table import table
from .table_image import table_image
from .utils import handle_not_tables


def handle_path(func):
    def inner_function(*args, **kwargs):
        try:
            return func(*args, **kwargs)
        except OSError as io:
            print(io)
            sys.exit()
        except OSError as exc:
            print(exc)
            sys.exit()
        except Exception as e:
            print(e)
            sys.exit()
        pass

    return inner_function


class autoCORPus:
    """ """

    @handle_path
    def __read_config(self, config_path):
        config_path = Path(config_path)
        with config_path.open("r") as f:
            ## TODO: validate config file here if possible
            content = json.load(f)
            return content["config"]

    @handle_path
    def __import_file(self, file_path):
        file_path = Path(file_path)
        with file_path.open("r") as f:
            return f.read(), file_path

    @handle_path
    def __handle_target_dir(self, target_dir):
        target_dir = Path(target_dir)
        if not target_dir.exists():
            target_dir.mkdir(parents=True)
        return

    def __validate_infile(self):
        pass

    def __soupify_infile(self, fpath):
        fpath = Path(fpath)
        try:
            with open(fpath, encoding="utf-8") as fp:
                soup = BeautifulSoup(fp.read(), "html.parser")
                for e in soup.find_all(
                    attrs={"style": ["display:none", "visibility:hidden"]}
                ):
                    e.extract()
                return soup
        except Exception as e:
            print(e)

    def __clean_text(self, result):
        """
        clean the main text body output of extract_text() further as follows:
            remove duplicated texts from each section (assuming the text from html file has hierarchy up to h3, i.e. no subsubsections);
            remove items with empty bodies

        Args:
            result: dict of the maintext


        Return:
            result: cleaned dict of the maintext
        """
        # Remove duplicated contents from the 'result' output of extract_text()

        # Identify unique section headings and the index of their first appearance
        idx_section = []
        section_headings = set([i["section_heading"] for i in result["paragraphs"]])

        for i in range(len(section_headings)):
            try:
                if idx_section[i + 1] - idx_section[i] <= 1:  # if only one subsection
                    continue
                idx_section_last = idx_section[i + 1]
            except IndexError:
                idx_section_last = len(result["paragraphs"])

            p = result["paragraphs"][idx_section[i] + 1]["body"]
            for idx_subsection in range(idx_section[i] + 1, idx_section_last):
                if (
                    result["paragraphs"][idx_subsection]["body"]
                    in result["paragraphs"][idx_section[i]]["body"]
                ):
                    result["paragraphs"][idx_section[i]]["body"] = result["paragraphs"][
                        idx_section[i]
                    ]["body"].replace(result["paragraphs"][idx_subsection]["body"], "")

                if (idx_section[i] + 1 != idx_subsection) and (
                    p in result["paragraphs"][idx_subsection]["body"]
                ):
                    result["paragraphs"][idx_subsection]["body"] = result["paragraphs"][
                        idx_subsection
                    ]["body"].replace(p, "")
            for idx_subsection in range(idx_section[i] + 1, idx_section_last):
                if (
                    result["paragraphs"][idx_subsection]["subsection_heading"]
                    == result["paragraphs"][idx_section[i]]["subsection_heading"]
                ):
                    result["paragraphs"][idx_section[i]]["subsection_heading"] = ""
        return result

    def __get_keywords(self, soup, config):
        if "keywords" in config:
            responses = handle_not_tables(config["keywords"], soup)
            responses = " ".join([x["node"].get_text() for x in responses])
            if not responses == "":
                keywordSection = {
                    "section_heading": "keywords",
                    "subsection_heading": "",
                    "body": responses,
                    "section_type": [
                        {"iao_name": "keywords section", "iao_id": "IAO:0000630"}
                    ],
                }
                return [keywordSection]
            return False

    def __get_title(self, soup, config):
        if "title" in config:
            titles = handle_not_tables(config["title"], soup)
            if len(titles) == 0:
                return ""
            else:
                return titles[0]["node"].get_text()
        else:
            return ""

    def __get_sections(self, soup, config):
        if "sections" in config:
            sections = handle_not_tables(config["sections"], soup)
            return sections
        return []

    def __extract_text(self, soup, config):
        """
        convert beautiful soup object into a python dict object with cleaned main text body

        Args:
            soup: BeautifulSoup object of html

        Return:
            result: dict of the maintext
        """
        result = {}

        # Tags of text body to be extracted are hard-coded as p (main text) and span (keywords and refs)

        # Extract title
        # try:
        # 	h1 = soup.find(config['title']['name'],
        # 	               config['title']['attrs']).get_text().strip('\n')
        # except:
        # 	h1 = ''
        result["title"] = self.__get_title(soup, config)
        maintext = (
            self.__get_keywords(soup, config)
            if self.__get_keywords(soup, config)
            else []
        )
        sections = self.__get_sections(soup, config)
        # sections = [x['node'] for x in sections]
        for sec in sections:
            maintext.extend(section(config, sec).to_dict())
        # filter out the sections which do not contain any info
        filteredText = []
        [filteredText.append(x) for x in maintext if x]
        uniqueText = []
        seen_text = []
        for text in filteredText:
            if text["body"] not in seen_text:
                seen_text.append(text["body"])
                uniqueText.append(text)

        result["paragraphs"] = self.__set_unknown_section_headings(uniqueText)

        return result

    def __set_unknown_section_headings(self, uniqueText):
        paper = {}
        for para in uniqueText:
            if para["section_heading"] != "keywords":
                paper[para["section_heading"]] = [
                    x["iao_name"] for x in para["section_type"]
                ]

        for text in uniqueText:
            if not text["section_heading"]:
                text["section_heading"] = "document part"
                text["section_type"] = [
                    {"iao_name": "document part", "iao_id": "IAO:0000314"}
                ]

        # uniqueText = [x for x in uniqueText if x['section_heading']]
        # mapping_dict_with_DAG = assgin_heading_by_DAG(paper)
        # for i, para in enumerate(uniqueText):
        #     if para['section_heading'] in mapping_dict_with_DAG.keys():
        #         if para['section_type'] == []:
        #             uniqueText[i]['section_type'] = mapping_dict_with_DAG[para['section_heading']]
        return uniqueText

    def __handle_html(self, file_path, config):
        """
        handles common HTML processing elements across main_text and linked_tables (creates soup and parses tables)
        :return: soup object
        """

        soup = self.__soupify_infile(file_path)
        if "tables" in config:
            if self.tables == {}:
                self.tables, self.empty_tables = table(
                    soup, config, file_path, self.base_dir
                ).to_dict()
            else:
                seenIDs = set()
                for tab in self.tables["documents"]:
                    if "." in tab["id"]:
                        seenIDs.add(tab["id"].split(".")[0])
                    else:
                        seenIDs.add(tab["id"])
                tmp_tables, tmp_empty = table(
                    soup, config, file_path, self.base_dir
                ).to_dict()
                for tabl in tmp_tables["documents"]:
                    if "." in tabl["id"]:
                        tabl_id = tabl["id"].split(".")[0]
                        tabl_pos = ".".join(tabl["id"].split(".")[1:])
                    else:
                        tabl_id = tabl["id"]
                        tabl_pos = None
                    if tabl_id in seenIDs:
                        tabl_id = str(len(seenIDs) + 1)
                        if tabl_pos:
                            tabl["id"] = f"{tabl_id}.{tabl_pos}"
                        else:
                            tabl["id"] = tabl_id
                    seenIDs.add(tabl_id)
                self.tables["documents"].extend(tmp_tables["documents"])
                self.empty_tables.extend(tmp_empty)
        return soup

    def __merge_table_data(self):
        if self.empty_tables == []:
            return
        if "documents" in self.tables:
            if self.tables["documents"] == []:
                return
            else:
                seenIDs = {}
                for i, table in enumerate(self.tables["documents"]):
                    if "id" in table:
                        seenIDs[str(i)] = f"Table {table['id']}."
                for table in self.empty_tables:
                    for seenID in seenIDs.keys():
                        if table["title"].startswith(seenIDs[seenID]):
                            if "title" in table and not table["title"] == "":
                                setNew = False
                                for passage in self.tables["documents"][int(seenID)][
                                    "passages"
                                ]:
                                    if (
                                        passage["infons"]["section_type"][0][
                                            "section_name"
                                        ]
                                        == "table_title"
                                    ):
                                        passage["text"] = table["title"]
                                        setNew = True
                                if not setNew:
                                    self.tables["documents"][int(seenID)][
                                        "passages"
                                    ].append(
                                        {
                                            "offset": 0,
                                            "infons": {
                                                "section_type": [
                                                    {
                                                        "section_name": "table_title",
                                                        "iao_name": "document title",
                                                        "iao_id": "IAO:0000305",
                                                    }
                                                ]
                                            },
                                            "text": table["title"],
                                        }
                                    )
                                pass
                            if "caption" in table and not table["caption"] == "":
                                setNew = False
                                for passage in self.tables["documents"][int(seenID)][
                                    "passages"
                                ]:
                                    if (
                                        passage["infons"]["section_type"][0][
                                            "section_name"
                                        ]
                                        == "table_caption"
                                    ):
                                        passage["text"] = table["caption"]
                                        setNew = True
                                if not setNew:
                                    self.tables["documents"][int(seenID)][
                                        "passages"
                                    ].append(
                                        {
                                            "offset": 0,
                                            "infons": {
                                                "section_type": [
                                                    {
                                                        "section_name": "table_caption",
                                                        "iao_name": "caption",
                                                        "iao_id": "IAO:0000304",
                                                    }
                                                ]
                                            },
                                            "text": table["caption"],
                                        }
                                    )
                                pass
                            if "footer" in table and not table["footer"] == "":
                                setNew = False
                                for passage in self.tables["documents"][int(seenID)][
                                    "passages"
                                ]:
                                    if (
                                        passage["infons"]["section_type"][0][
                                            "section_name"
                                        ]
                                        == "table_footer"
                                    ):
                                        passage["text"] = table["footer"]
                                        setNew = True
                                if not setNew:
                                    self.tables["documents"][int(seenID)][
                                        "passages"
                                    ].append(
                                        {
                                            "offset": 0,
                                            "infons": {
                                                "section_type": [
                                                    {
                                                        "section_name": "table_footer",
                                                        "iao_name": "caption",
                                                        "iao_id": "IAO:0000304",
                                                    }
                                                ]
                                            },
                                            "text": table["footer"],
                                        }
                                    )
        else:
            return

    def __init__(
        self,
        config_path,
        base_dir=None,
        main_text=None,
        linked_tables=None,
        table_images=None,
        associated_data_path=None,
        trainedData=None,
    ):
        """

        :param config_path: path to the config file to be used
        :param file_path: path to the main text of the article (HTML files only)
        :param linked_tables: list of linked table file paths to be included in this run (HTML files only)
        :param table_images: list of table image file paths to be included in this run (JPEG or PNG files only)
        :param associated_data_path: this still needs sorting
        """
        # handle common
        config = self.__read_config(config_path)
        self.base_dir = base_dir
        self.file_path = main_text
        self.main_text = {}
        self.empty_tables = {}
        self.tables = {}
        self.abbreviations = {}
        self.has_tables = False

        # handle main_text
        if self.file_path:
            soup = self.__handle_html(self.file_path, config)
            self.main_text = self.__extract_text(soup, config)
            try:
                self.abbreviations = abbreviations(
                    self.main_text, soup, config, self.file_path
                ).to_dict()
            except Exception as e:
                print(e)
        if linked_tables:
            for table_file in linked_tables:
                soup = self.__handle_html(table_file, config)
        if table_images:
            self.tables = table_image(
                table_images, self.base_dir, trainedData=trainedData
            ).to_dict()
        self.__merge_table_data()
        if "documents" in self.tables and not self.tables["documents"] == []:
            self.has_tables = True

    def to_bioc(self):
        return BiocFormatter(self).to_dict()

    def main_text_to_bioc_json(self, indent=2):
        return BiocFormatter(self).to_json(indent)

    def main_text_to_bioc_xml(self):
        collection = biocjson.loads(BiocFormatter(self).to_json(2))
        return biocxml.dumps(collection)

    def tables_to_bioc_json(self, indent=2):
        return json.dumps(self.tables, ensure_ascii=False, indent=indent)

    def abbreviations_to_bioc_json(self, indent=2):
        return json.dumps(self.abbreviations, ensure_ascii=False, indent=indent)

    def to_json(self, indent=2):
        return json.dumps(self.to_dict(), ensure_ascii=False, indent=indent)

    def to_dict(self):
        return {
            "main_text": self.main_text,
            "abbreviations": self.abbreviations,
            "tables": self.tables,
        }


if __name__ == "__main__":
    parser = argparse.ArgumentParser()
    parser.add_argument(
        "-f", "--filepath", type=str, help="filepath of html file to be processed"
    )
    parser.add_argument(
        "-t", "--target_dir", type=str, help="target directory for output"
    )
    parser.add_argument(
        "-c", "--config", type=str, help="filepath for configuration JSON file"
    )
    parser.add_argument(
        "-d", "--config_dir", type=str, help="directory of configuration JSON files"
    )
    parser.add_argument(
        "-a", "--associated_data", type=str, help="directory of associated data"
    )
    args = parser.parse_args()
    filepath = args.filepath
    target_dir = args.target_dir
    config_path = args.config

    autoCORPus(config_path, filepath, target_dir).to_file(target_dir)



import logging
from collections import Counter, defaultdict
from datetime import datetime
from pathlib import Path

import regex as re2


class abbreviations:
    def __yield_lines_from_doc(self, doc_text):
        for line in doc_text.split("."):
            yield line.strip()

    def __conditions(self, candidate):
        r"""
        Based on Schwartz&Hearst

        2 <= len(str) <= 10
        len(tokens) <= 2
        re.search(r'\p{L}', str)
        str[0].isalnum()

        and extra:
        if it matches (\p{L}\.?\s?){2,}
        it is a good candidate.

        :param candidate: candidate abbreviation
        :return: True if this is a good candidate
        """
        viable = True
        if re2.match(r"(\p{L}\.?\s?){2,}", candidate.lstrip()):
            viable = True
        if len(candidate) < 2 or len(candidate) > 10:
            viable = False
        if len(candidate.split()) > 2:
            viable = False
        if candidate.islower():  # customize function discard all lower case candidate
            viable = False
        if not re2.search(r"\p{L}", candidate):  # \p{L} = All Unicode letter
            viable = False
        if not candidate[0].isalnum():
            viable = False

        return viable

    def __best_candidates(self, sentence):
        """
        :param sentence: line read from input file
        :return: a Candidate iterator
        """

        if "(" in sentence:
            # Check some things first
            if sentence.count("(") != sentence.count(")"):
                raise ValueError(f"Unbalanced parentheses: {sentence}")

            if sentence.find("(") > sentence.find(")"):
                raise ValueError(f"First parentheses is right: {sentence}")

            close_index = -1
            while 1:
                # Look for open parenthesis. Need leading whitespace to avoid matching mathematical and chemical formulae
                open_index = sentence.find(" (", close_index + 1)

                if open_index == -1:
                    break

                # Advance beyond whitespace
                open_index += 1

                # Look for closing parentheses
                close_index = open_index + 1
                open_count = 1
                skip = False
                while open_count:
                    try:
                        char = sentence[close_index]
                    except IndexError:
                        # We found an opening bracket but no associated closing bracket
                        # Skip the opening bracket
                        skip = True
                        break
                    if char == "(":
                        open_count += 1
                    elif char in [")", ";", ":"]:
                        open_count -= 1
                    close_index += 1

                if skip:
                    close_index = open_index + 1
                    continue

                # Output if conditions are met
                start = open_index + 1
                stop = close_index - 1
                candidate = sentence[start:stop]

                # Take into account whitespace that should be removed
                start = start + len(candidate) - len(candidate.lstrip())
                stop = stop - len(candidate) + len(candidate.rstrip())
                candidate = sentence[start:stop]
                # print (candidate)

                if self.__conditions(candidate):
                    new_candidate = Candidate(candidate)
                    new_candidate.set_position(start, stop)
                    yield new_candidate

    # elif LF_in_parentheses:

    def __get_definition(self, candidate, sentence):
        """
        Takes a candidate and a sentence and returns the definition candidate.

        The definition candidate is the set of tokens (in front of the candidate)
        that starts with a token starting with the first character of the candidate

        :param candidate: candidate abbreviation
        :param sentence: current sentence (single line from input file)
        :return: candidate definition for this abbreviation
        """
        # Take the tokens in front of the candidate
        tokens = re2.split(r"[\s\-]+", sentence[: candidate.start - 2].lower())
        # the char that we are looking for
        key = candidate[0].lower()

        # Count the number of tokens that start with the same character as the candidate
        first_chars = [t[0] for t in filter(None, tokens)]

        definition_freq = first_chars.count(key)
        candidate_freq = candidate.lower().count(key)

        # Look for the list of tokens in front of candidate that
        # have a sufficient number of tokens starting with key
        if candidate_freq <= definition_freq:
            # we should at least have a good number of starts
            count = 0
            start = 0
            start_index = len(first_chars) - 1
            while count < candidate_freq:
                if abs(start) > len(first_chars):
                    raise ValueError(f"candidate {candidate} not found")
                start -= 1
                # Look up key in the definition
                try:
                    start_index = first_chars.index(key, len(first_chars) + start)
                except ValueError:
                    pass

                # Count the number of keys in definition
                count = first_chars[start_index:].count(key)

            # We found enough keys in the definition so return the definition as a definition candidate
            start = len(" ".join(tokens[:start_index]))
            stop = candidate.start - 1
            candidate = sentence[start:stop]

            # Remove whitespace
            start = start + len(candidate) - len(candidate.lstrip())
            stop = stop - len(candidate) + len(candidate.rstrip())
            candidate = sentence[start:stop]

            new_candidate = Candidate(candidate)
            new_candidate.set_position(start, stop)
            return new_candidate

        else:
            raise ValueError(
                "There are less keys in the tokens in front of candidate than there are in the candidate"
            )

    def __select_definition(self, definition, abbrev):
        """
        Takes a definition candidate and an abbreviation candidate
        and returns True if the chars in the abbreviation occur in the definition

        Based on
        A simple algorithm for identifying abbreviation definitions in biomedical texts, Schwartz & Hearst
        :param definition: candidate definition
        :param abbrev: candidate abbreviation
        :return:
        """

        if len(definition) < len(abbrev):
            raise ValueError("Abbreviation is longer than definition")

        if abbrev in definition.split():
            raise ValueError("Abbreviation is full word of definition")

        s_index = -1
        l_index = -1

        while 1:
            try:
                long_char = definition[l_index].lower()
            except IndexError:
                raise

            short_char = abbrev[s_index].lower()

            if not short_char.isalnum():
                s_index -= 1

            if s_index == -1 * len(abbrev):
                if short_char == long_char:
                    if (
                        l_index == -1 * len(definition)
                        or not definition[l_index - 1].isalnum()
                    ):
                        break
                    else:
                        l_index -= 1
                else:
                    l_index -= 1
                    if l_index == -1 * (len(definition) + 1):
                        raise ValueError(
                            f"definition {abbrev} was not found in {definition}"
                        )

            else:
                if short_char == long_char:
                    s_index -= 1
                    l_index -= 1
                else:
                    l_index -= 1

        new_candidate = Candidate(definition[l_index : len(definition)])
        new_candidate.set_position(definition.start, definition.stop)
        definition = new_candidate

        tokens = len(definition.split())
        length = len(abbrev)

        if tokens > min([length + 5, length * 2]):
            raise ValueError("did not meet min(|A|+5, |A|*2) constraint")

        # Do not return definitions that contain unbalanced parentheses
        if definition.count("(") != definition.count(")"):
            raise ValueError("Unbalanced parentheses not allowed in a definition")

        return definition

    def __extract_abbreviation_definition_pairs(
        self,
        doc_text=None,
        most_common_definition=False,
        first_definition=False,
        all_definition=True,
    ):
        abbrev_map = dict()
        list_abbrev_map = defaultdict(list)
        counter_abbrev_map = dict()
        omit = 0
        written = 0
        sentence_iterator = enumerate(self.__yield_lines_from_doc(doc_text))

        collect_definitions = False
        if most_common_definition or first_definition or all_definition:
            collect_definitions = True

        for i, sentence in sentence_iterator:
            # Remove any quotes around potential candidate terms
            clean_sentence = re2.sub(
                r'([(])[\'"\p{Pi}]|[\'"\p{Pf}]([);:])', r"\1\2", sentence
            )
            try:
                for candidate in self.__best_candidates(clean_sentence):
                    try:
                        definition = self.__get_definition(candidate, clean_sentence)
                    except (ValueError, IndexError) as e:
                        self.log.debug(
                            f"{i} Omitting candidate {candidate}. Reason: {e.args[0]}"
                        )
                        omit += 1
                    else:
                        try:
                            definition = self.__select_definition(definition, candidate)
                        except (ValueError, IndexError) as e:
                            self.log.debug(
                                f"{i} Omitting definition {definition} for candidate {candidate}. Reason: {e.args[0]}"
                            )
                            omit += 1
                        else:
                            # Either append the current definition to the list of previous definitions ...
                            if collect_definitions:
                                list_abbrev_map[candidate].append(definition)
                            else:
                                # Or update the abbreviations map with the current definition
                                abbrev_map[candidate] = definition
                            written += 1
            except (ValueError, IndexError) as e:
                self.log.debug(f"{i} Error processing sentence {sentence}: {e.args[0]}")
        self.log.debug(f"{written} abbreviations detected and kept ({omit} omitted)")

        # Return most common definition for each term
        if collect_definitions:
            if most_common_definition:
                # Return the most common definition for each term
                for k, v in list_abbrev_map.items():
                    counter_abbrev_map[k] = Counter(v).most_common(1)[0][0]
            elif first_definition:
                # Return the first definition for each term
                for k, v in list_abbrev_map.items():
                    counter_abbrev_map[k] = v
            elif all_definition:
                for k, v in list_abbrev_map.items():
                    counter_abbrev_map[k] = v
            return counter_abbrev_map

        # Or return the last encountered definition for each term
        return abbrev_map

    def __extract_abbreviation(self, main_text):
        pairs = self.__extract_abbreviation_definition_pairs(
            doc_text=main_text, most_common_definition=True
        )

        return pairs

    def __listToDict(self, lst):
        op = {lst[i]: lst[i + 1] for i in range(0, len(lst), 2)}
        return op

    def __abbre_table_to_dict(self, t):
        abbre_list = []
        rows = t.findAll("tr")
        for i in rows:
            elements = i.findAll(["td", "th"])
            vals = [j.get_text() for j in elements]
            if len(vals) > 1:
                abbre_list += vals
        abbre_dict = self.__listToDict(abbre_list)
        return abbre_dict

    def __abbre_list_to_dict(self, t):
        SF = t.findAll("dt")
        SF_list = [SF_word.get_text() for SF_word in SF]
        LF = t.findAll("dd")
        LF_list = [LF_word.get_text() for LF_word in LF]
        abbre_dict = dict(zip(SF_list, LF_list))
        return abbre_dict

    def __get_abbre_plain_text(self, soup_og):
        abbre_text = soup_og.get_text()
        abbre_list = abbre_text.split(";")
        list_lenth = len(abbre_list)
        return abbre_list, list_lenth

    def __get_abbre_dict_given_by_author(self, soup_og):
        header = soup_og.find_all("h2", recursive=True)
        abbre_dict = {}
        for number, element in enumerate(header):
            if re2.search("abbreviation", element.get_text(), re2.IGNORECASE):
                nearest_down_tag = element.next_element
                while nearest_down_tag:
                    tag_name = nearest_down_tag.name

                    # when abbre is table
                    if tag_name == "table":
                        abbre_dict = self.__abbre_table_to_dict(nearest_down_tag)
                        break

                    # when abbre is list
                    elif tag_name == "dl":
                        abbre_dict = self.__abbre_list_to_dict(nearest_down_tag)
                        break

                    # when abbre is plain text
                    elif tag_name == "p":
                        abbre_list, list_lenth = self.__get_abbre_plain_text(
                            nearest_down_tag
                        )
                        if list_lenth <= 2:
                            nearest_down_tag = nearest_down_tag.next_element
                            continue
                        else:
                            for abbre_pair in abbre_list:
                                if len(abbre_pair.split(":")) == 2:
                                    abbre_dict.update(
                                        {
                                            abbre_pair.split(":")[0]: abbre_pair.split(
                                                ":"
                                            )[1]
                                        }
                                    )
                                elif len(abbre_pair.split(",")) == 2:
                                    abbre_dict.update(
                                        {
                                            abbre_pair.split(",")[0]: abbre_pair.split(
                                                ","
                                            )[1]
                                        }
                                    )
                                elif len(abbre_pair.split(" ")) == 2:
                                    abbre_dict.update(
                                        {
                                            abbre_pair.split(" ")[0]: abbre_pair.split(
                                                " "
                                            )[1]
                                        }
                                    )
                            break

                    # search until next h2
                    elif tag_name == "h2":
                        break
                    else:
                        nearest_down_tag = nearest_down_tag.next_element
        return abbre_dict

    def __get_abbreviations(self, main_text, soup, config):
        paragraphs = main_text["paragraphs"]
        all_abbreviations = {}
        for paragraph in paragraphs:
            maintext = paragraph["body"]
            pairs = self.__extract_abbreviation(maintext)
            all_abbreviations.update(pairs)
        author_provided_abbreviations = self.__get_abbre_dict_given_by_author(soup)

        abbrev_json = {}

        for key in author_provided_abbreviations.keys():
            abbrev_json[key] = {
                author_provided_abbreviations[key]: ["abbreviations section"]
            }
        for key in all_abbreviations:
            if key in abbrev_json:
                if all_abbreviations[key] in abbrev_json[key].keys():
                    abbrev_json[key][all_abbreviations[key]].append("fulltext")
                else:
                    abbrev_json[key][all_abbreviations[key]] = ["fulltext"]
            else:
                abbrev_json[key] = {all_abbreviations[key]: ["fulltext"]}

        # abbrev_json['abbreviations_section'] = author_provided_abbreviations
        # abbrev_json['fulltext_algorithm'] = all_abbreviations
        return abbrev_json

    def __biocify_abbreviations(self, abbreviations, file_path):
        template = {
            "source": "Auto-CORPus (abbreviations)",
            # "inputfile": file_path,
            "date": f'{datetime.today().strftime("%Y%m%d")}',
            "key": "autocorpus_abbreviations.key",
            "documents": [
                {
                    "id": Path(file_path).name.split(".")[0],
                    "inputfile": file_path,
                    "passages": [],
                }
            ],
        }
        passages = template["documents"][0]["passages"]
        for short in abbreviations.keys():
            counter = 1
            shortTemplate = {"text_short": short}
            for long in abbreviations[short].keys():
                shortTemplate[f"text_long_{counter}"] = long.replace("\n", " ")
                shortTemplate[f"extraction_algorithm_{counter}"] = ", ".join(
                    abbreviations[short][long]
                )
                counter += 1
            passages.append(shortTemplate)
        return template

    def __init__(self, main_text, soup, config, file_path):
        logging.basicConfig(
            format="%(asctime)s : %(levelname)s : %(message)s", level=logging.INFO
        )
        self.log = logging.getLogger(__name__)

        self.abbreviations = self.__biocify_abbreviations(
            self.__get_abbreviations(main_text, soup, config), file_path
        )
        pass

    def to_dict(self):
        return self.abbreviations


class Candidate(str):
    def __init__(self, value):
        super().__init__()
        self.start = 0
        self.stop = 0

    def set_position(self, start, stop):
        self.start = start
        self.stop = stop



import re
import unicodedata
from importlib import resources
from pathlib import Path

import bs4
import networkx as nx
from bs4 import NavigableString
from lxml import etree
from lxml.html.soupparser import fromstring


def get_files(base_dir, pattern=r"(.*).html"):
    """
    recursively retrieve all PMC.html files from the directory

    Args:
        base_dir: base directory
        pattern: file name filter REGEX pattern (default *.html)

    Return:
        file_list: a list of filepath
    """
    file_list = []
    base_dir = Path(base_dir)
    for item in base_dir.iterdir():
        abs_path = item.resolve()
        if abs_path.is_file() and re.match(pattern, str(abs_path)):
            file_list.append(str(abs_path))
        elif abs_path.is_dir() and "ipynb_checkpoints" not in str(abs_path):
            file_list += get_files(abs_path, pattern)
    return file_list


def process_supsub(soup):
    """
    add underscore (_) before all superscript or subscript text

    Args:
        soup: BeautifulSoup object of html

    """
    # for sup in soup.find_all(['sup', 'sub']):
    for sup in soup.find_all("sub"):
        s = sup.get_text()
        if sup.string is None:
            sup.extract()
        elif re.match("[_-]", s):
            sup.string.replace_with(f"{s} ")
        else:
            sup.string.replace_with(f"_{s} ")
    return soup


def process_em(soup):
    """
    remove all emphasized text
    No it doesn't, it just adds a space to it

    Args:
        soup: BeautifulSoup object of html

    """
    for em in soup.find_all("em"):
        s = em.get_text()
        if em.string is None:
            em.extract()
        else:
            em.string.replace_with(f"{s} ")
    return soup


def read_mapping_file():
    mapping_dict = {}
    mapping_path = resources.files("autocorpus.IAO_dicts") / "IAO_FINAL_MAPPING.txt"
    with mapping_path.open(encoding="utf-8") as f:
        lines = f.readlines()
        for line in lines:
            heading = line.split("\t")[0].lower().strip("\n")
            IAO_term = line.split("\t")[1].lower().strip("\n")
            if IAO_term != "":
                if "/" in IAO_term:
                    IAO_term_1 = IAO_term.split("/")[0].strip(" ")
                    IAO_term_2 = IAO_term.split("/")[1].strip(" ")
                    if IAO_term_1 in mapping_dict.keys():
                        mapping_dict[IAO_term_1].append(heading)
                    else:
                        mapping_dict.update({IAO_term_1: [heading]})

                    if IAO_term_2 in mapping_dict.keys():
                        mapping_dict[IAO_term_2].append(heading)
                    else:
                        mapping_dict.update({IAO_term_2: [heading]})

                else:
                    if IAO_term in mapping_dict.keys():
                        mapping_dict[IAO_term].append(heading)
                    else:
                        mapping_dict.update({IAO_term: [heading]})
    return mapping_dict


def read_IAO_term_to_ID_file():
    IAO_term_to_no_dict = {}
    ID_path = resources.files("autocorpus.IAO_dicts") / "IAO_term_to_ID.txt"
    with ID_path.open(encoding="utf-8") as f:
        lines = f.readlines()
        for line in lines:
            IAO_term = line.split("\t")[0]
            IAO_no = line.split("\t")[1].strip("\n")
            IAO_term_to_no_dict.update({IAO_term: IAO_no})
    return IAO_term_to_no_dict


def config_anchors(value):
    if not value.startswith("^"):
        value = f"^{value}"
    if not value.endswith("$"):
        value = f"{value}$"
    return value


def config_attr_block(block):
    ret = {}
    for key in block:
        if isinstance(block[key], list):
            ret[key] = [re.compile(config_anchors(x)) for x in block[key]]
        elif isinstance(block[key], str):
            ret[key] = re.compile(config_anchors(block[key]))
    return ret


def config_attrs(attrs):
    ret = []
    if isinstance(attrs, list):
        for attr in attrs:
            ret.extend(config_attr_block(attr))
    elif isinstance(attrs, dict):
        ret = config_attr_block(attrs)
    else:
        quit(f"{attrs} must be a dict or a list of dicts")
    return ret


def config_tags(tags):
    ret = []
    if isinstance(tags, list):
        for tag in tags:
            if isinstance(tag, str):
                ret.append(re.compile(config_anchors(tag)))
            else:
                quit(f"{tags} must be a string or list of strings")
    elif isinstance(tags, str):
        ret.append(re.compile(config_anchors(tags)))
    else:
        quit(f"{tags} must be a string or list of strings")
    return ret


def parse_configs(definition):
    bsAttrs = {"name": [], "attrs": [], "xpath": []}
    if "tag" in definition:
        bsAttrs["name"] = config_tags(definition["tag"])
    if "attrs" in definition:
        bsAttrs["attrs"] = config_attrs(definition["attrs"])
    if "xpath" in definition:
        bsAttrs["xpath"] = definition["xpath"]
    return bsAttrs


def handle_defined_by(config, soup):
    """

    :param config: config file section used to parse
    :param soup: soup section to parse
    :return:
    list of objects, each object being a matching node. Object of the form:
            {
                    node: bs4Object,
                    data:{
                                    key: [values]
                            }
            }
    node is a bs4 object of a single result derived from bs4.find_all()
    data is an object where the results from the config "data" sections is housed. The key is the name of the data
    section and the values are all matches found within any of the main matches which match the current data section
    definition. The values is the response you get from get_text() on any found nodes, not the nodes themselves.
    """
    if "defined-by" not in config:
        quit(f"{config} does not contain the required 'defined-by' key.")
    matches = []
    seen_text = []
    for definition in config["defined-by"]:
        bsAttrs = parse_configs(definition)
        new_matches = []
        if bsAttrs["name"] or bsAttrs["attrs"]:
            new_matches = soup.find_all(bsAttrs["name"], bsAttrs["attrs"])
            if new_matches:
                new_matches = [x for x in new_matches if x.text]
        if "xpath" in bsAttrs:
            if isinstance(bsAttrs["xpath"], list):
                for path in bsAttrs["xpath"]:
                    xpath_matches = fromstring(str(soup)).xpath(path)
                    if xpath_matches:
                        for new_match in xpath_matches:
                            new_match = bs4.BeautifulSoup(
                                etree.tostring(
                                    new_match, encoding="unicode", method="html"
                                ),
                                "html.parser",
                            )
                            if new_match.text.strip():
                                new_matches.extend(new_match)
            else:
                xpath_matches = fromstring(str(soup)).xpath(bsAttrs["xpath"])
                if xpath_matches:
                    for new_match in xpath_matches:
                        new_match = bs4.BeautifulSoup(
                            etree.tostring(
                                new_match, encoding="unicode", method="html"
                            ),
                            "html.parser",
                        )
                        if new_match.text.strip():
                            new_matches.extend(new_match)
        for match in new_matches:
            if type(match) is not NavigableString:
                matched_text = match.get_text()
            if matched_text in seen_text:
                continue
            else:
                seen_text.append(matched_text)
                matches.append(match)
    return matches


def handle_not_tables(config, soup):
    responses = []
    matches = handle_defined_by(config, soup)
    if "data" in config:
        for match in matches:
            responseAddition = {"node": match}
            for ele in config["data"]:
                seen_text = set()
                for definition in config["data"][ele]:
                    bsAttrs = parse_configs(definition)
                    newMatches = match.find_all(bsAttrs["name"], bsAttrs["attrs"])
                    if newMatches:
                        responseAddition[ele] = []
                    for newMatch in newMatches:
                        if newMatch.get_text() in seen_text:
                            continue
                        else:
                            responseAddition[ele].append(newMatch.get_text())
            responses.append(responseAddition)
    else:
        for match in matches:
            responseAddition = {"node": match}
            responses.append(responseAddition)
    return responses


def get_data_element_node(config, soup):
    config = {"defined-by": config}
    return handle_defined_by(config, soup)


def navigate_contents(item):
    value = ""
    if isinstance(item, bs4.element.NavigableString):
        value += unicodedata.normalize("NFKD", item)
    if isinstance(item, bs4.element.Tag):
        if item.name == "sup" or item.name == "sub":
            value += "<" + item.name + ">"
            for childItem in item.contents:
                value += navigate_contents(childItem)
            value += "</" + item.name + ">"
        else:
            for childItem in item.contents:
                value += navigate_contents(childItem)
    return value


def handle_tables(config, soup):
    responses = []
    matches = handle_defined_by(config, soup)
    textData = ["caption", "title", "footer"]
    if "data" in config:
        for match in matches:
            responseAddition = {"node": match, "title": "", "footer": "", "caption": ""}
            for ele in config["data"]:
                if ele in textData:
                    seen_text = set()
                    for definition in config["data"][ele]:
                        bsAttrs = parse_configs(definition)
                        newMatches = match.find_all(bsAttrs["name"], bsAttrs["attrs"])
                        if newMatches:
                            responseAddition[ele] = []
                        for newMatch in newMatches:
                            if newMatch.get_text() in seen_text:
                                continue
                            else:
                                value = ""
                                for item in newMatch.contents:
                                    value += navigate_contents(item)
                                    # clean the cell
                                value = value.strip().replace("\u2009", " ")
                                value = re.sub("<\\/?span[^>\n]*>?|<hr\\/>?", "", value)
                                value = re.sub("\\n", "", value)
                                responseAddition[ele].append(value)
            responses.append(responseAddition)
    else:
        for match in matches:
            responseAddition = {"node": match}
            responses.append(responseAddition)
    return responses


def assgin_heading_by_DAG(paper):
    G = nx.read_graphml(resources.files("autocorpus") / "DAG_model.graphml")
    new_mapping_dict = {}
    mapping_dict_with_DAG = {}
    IAO_term_to_no_dict = read_IAO_term_to_ID_file()
    for i, heading in enumerate(paper.keys()):
        if paper[heading] == []:
            previous_mapped_heading_found = False
            i2 = 1
            while not previous_mapped_heading_found:
                if i - i2 > len(list(paper.keys())):
                    previous_mapped_heading_found = True
                    previous_section = "Start of the article"
                else:
                    previous_heading = list(paper.keys())[i - i2]
                    if paper[previous_heading] != []:
                        previous_mapped_heading_found = True
                        previous_section = paper[previous_heading]
                    else:
                        i2 += 1

            next_mapped_heading_found = False
            i2 = 1
            while not next_mapped_heading_found:
                if i + i2 >= len(list(paper.keys())):
                    next_mapped_heading_found = True
                    next_section = "End of the article"
                else:
                    next_heading = list(paper.keys())[i + i2]
                    if paper[next_heading] != []:
                        next_mapped_heading_found = True
                        next_section = paper[next_heading]
                    else:
                        i2 += 1

            if (
                previous_section != "Start of the article"
                and next_section != "End of the article"
            ):
                try:
                    paths = nx.all_shortest_paths(
                        G,
                        paper[previous_heading][-1],
                        paper[next_heading][0],
                        weight="cost",
                    )
                    for path in paths:
                        if len(path) <= 2:
                            mapping_dict_with_DAG.update({heading: [path[0]]})
                        if len(path) > 2:
                            mapping_dict_with_DAG.update({heading: path[1:-1]})
                except Exception:
                    new_target = paper[list(paper.keys())[i + i2 + 1]][0]
                    paths = nx.all_shortest_paths(
                        G, paper[previous_heading][-1], new_target, weight="cost"
                    )
                    for path in paths:
                        if len(path) == 2:
                            mapping_dict_with_DAG.update({heading: [path[0]]})
                        if len(path) > 2:
                            mapping_dict_with_DAG.update({heading: path[1:-1]})

            if next_section == "End of the article":
                mapping_dict_with_DAG.update({heading: [previous_section[-1]]})

            for heading in mapping_dict_with_DAG.keys():
                newSecType = []
                for secType in mapping_dict_with_DAG[heading]:
                    if secType in IAO_term_to_no_dict.keys():
                        mapping_result_ID_version = IAO_term_to_no_dict[secType]
                    else:
                        mapping_result_ID_version = ""
                    newSecType.append(
                        {"iao_name": secType, "iao_id": mapping_result_ID_version}
                    )

                new_mapping_dict[heading] = newSecType
    return new_mapping_dict



import re


class references:
    #
    # def __get_section_header(self, soup_section):
    # 	h2 = ""
    # 	h2_select_tag = self.config['heading']['name']
    # 	h2_select_tag += ''.join(['[{}*={}]'.format(k, self.config['heading']['attrs'][k])
    # 	                          for k in self.config['heading']['attrs'] if self.config['heading']['attrs'][k]])
    # 	_h2 = soup_section.select(h2_select_tag)
    # 	if _h2:
    # 		h2 = _h2[0].get_text().strip('\n')
    # 	return h2
    # 	pass
    #
    # def __get_subsection_header(self, soup_section):
    # 	h3_select_tag = self.config['heading2']['name']
    # 	h3_select_tag += ''.join(['[{}*={}]'.format(k, self.config['heading2']['attrs'][k])
    # 	                          for k in self.config['heading2']['attrs'] if self.config['heading2']['attrs'][k]])
    # 	h3 = soup_section.select(h3_select_tag)
    # 	if h3:
    # 		h3 = h3[0].get_text().strip('\n')
    # 	else:
    # 		h3 = ''
    # 	return h3
    # 	pass

    def __create_reference_block(self, reference):
        text = reference["node"].get_text().replace("Go to:", "").replace("\n", "")
        text = re.sub(r"\s{2,}", " ", text)
        refSection = {
            "section_heading": self.section_heading,
            "subsection_heading": "",
            "body": text,
            "section_type": [
                {"iao_name": "references section", "iao_id": "IAO:0000320"}
            ],
        }

        for subsec in reference:
            if subsec == "node":
                continue
            refSection[subsec] = ". ".join(reference[subsec])

        return refSection

    def __init__(self, soup, config, section_heading):
        self.config = config
        self.section_heading = section_heading

        self.reference = self.__create_reference_block(soup)

    def to_dict(self):
        return self.reference



import re

import nltk
from fuzzywuzzy import fuzz

from .references import references
from .utils import handle_not_tables, read_IAO_term_to_ID_file, read_mapping_file


class section:
    # def __get_section_header(self, soup_section):
    #
    # 	# if "sectionsNew" in config:
    # 	h2 = ""
    # 	h2_select_tag = self.config['heading']['name']
    # 	h2_select_tag += ''.join(['[{}*={}]'.format(k, self.config['heading']['attrs'][k])
    # 	                          for k in self.config['heading']['attrs'] if self.config['heading']['attrs'][k]])
    # 	_h2 = soup_section.select(h2_select_tag)
    # 	if _h2:
    # 		h2 = _h2[0].get_text().strip('\n')
    # 	return h2

    # def __get_subsection_header(self, soup_section):
    # 	h3_select_tag = self.config['heading2']['name']
    # 	h3_select_tag += ''.join(['[{}*={}]'.format(k, self.config['heading2']['attrs'][k])
    #                           for k in self.config['heading2']['attrs'] if self.config['heading2']['attrs'][k]])
    # 	h3 = soup_section.select(h3_select_tag)
    # 	if h3:
    # 		h3 = h3[0].get_text().strip('\n')
    # 	else:
    # 		h3 = ''
    # 	return h3

    def __add_paragraph(self, body):
        self.paragraphs.append(
            {
                "section_heading": self.section_heading,
                "subsection_heading": self.subheader,
                "body": body,
                "section_type": self.section_type,
            }
        )

    def __navigate_children(self, soup_section, all_sub_sections, filtered_paragraphs):
        if soup_section in filtered_paragraphs:
            if soup_section.previous_sibling and soup_section.previous_sibling.name in (
                "h3",
                "h4",
                "h5",
            ):
                self.subheader = soup_section.previous_sibling.get_text()
            self.__add_paragraph(soup_section.get_text())
            return
        for subsec in all_sub_sections:
            if subsec["node"] == soup_section:
                self.subheader = (
                    subsec["headers"][0]
                    if "headers" in subsec and not subsec["headers"] == ""
                    else ""
                )
        # elif soup_section in subsecNodes:
        # 	self.subheader = self.__get_subsection_header(soup_section)
        try:
            children = soup_section.findChildren(recursive=False)
        except Exception as e:
            print(e)
            children = []
        for child in children:
            self.__navigate_children(child, all_sub_sections, filtered_paragraphs)

    def __get_abbreviations(self, soup_section):
        if "abbreviations-table" in self.config:
            try:
                abbreviations_tables = handle_not_tables(
                    self.config["abbreviations-table"], soup_section
                )
                abbreviations_tables = abbreviations_tables[0]["node"]
                abbreviations = {}
                for tr in abbreviations_tables.find_all("tr"):
                    short_form, long_form = (td.get_text() for td in tr.find_all("td"))
                    abbreviations[short_form] = long_form
            except Exception:
                abbreviations = {}
            self.__add_paragraph(str(abbreviations))

    def __set_IAO(self):
        mapping_dict = read_mapping_file()
        tokenized_section_heading = nltk.wordpunct_tokenize(self.section_heading)
        text = nltk.Text(tokenized_section_heading)
        ## this .isalpha() should probably be removed as it;s stripping out &
        # words = [w.lower() for w in text if w.isalpha()]
        words = [w.lower() for w in text]
        h2_tmp = " ".join(word for word in words)

        # TODO: check for best match, not the first
        if h2_tmp != "":
            if any(x in h2_tmp for x in [" and ", "&", "/"]):
                mapping_result = []
                h2_parts = re.split(r" and |\s?/\s?|\s?&\s?", h2_tmp)
                for h2_part in h2_parts:
                    h2_part = re.sub(r"^\d*\s?[\(\.]]?\s?", "", h2_part)
                    pass
                    for IAO_term, heading_list in mapping_dict.items():
                        if any(
                            [
                                fuzz.ratio(h2_part, heading) >= 80
                                for heading in heading_list
                            ]
                        ):
                            mapping_result.append(self.__add_IAO(IAO_term))
                            break

            else:
                for IAO_term, heading_list in mapping_dict.items():
                    h2_tmp = re.sub(r"^\d*\s?[\(\.]]?\s?", "", h2_tmp)
                    if any(
                        [fuzz.ratio(h2_tmp, heading) > 80 for heading in heading_list]
                    ):
                        mapping_result = [self.__add_IAO(IAO_term)]
                        break
                    else:
                        mapping_result = []
        else:
            mapping_result = []
        self.section_type = mapping_result

    def __add_IAO(self, IAO_term):
        paper = {}
        IAO_term = IAO_term
        paper.update({self.section_heading: IAO_term})
        # mapping_dict_with_DAG = assgin_heading_by_DAG(paper)
        #
        # if self.section_heading in mapping_dict_with_DAG.keys():
        # 	IAO_term = mapping_dict_with_DAG[self.section_heading]

        # map IAO terms to IAO IDs
        IAO_term_to_no_dict = read_IAO_term_to_ID_file()
        if IAO_term in IAO_term_to_no_dict.keys():
            mapping_result_ID_version = IAO_term_to_no_dict[IAO_term]
        else:
            mapping_result_ID_version = ""
        return {"iao_name": IAO_term, "iao_id": mapping_result_ID_version}

    def __get_section(self, soup_section):
        all_subSections = handle_not_tables(self.config["sub-sections"], soup_section)
        all_paragraphs = handle_not_tables(self.config["paragraphs"], soup_section)
        all_paragraphs = [x["node"] for x in all_paragraphs]
        all_tables = handle_not_tables(self.config["tables"], soup_section)
        all_tables = [x["node"] for x in all_tables]
        all_figures = handle_not_tables(self.config["figures"], soup_section)
        all_figures = [x["node"] for x in all_figures]
        # all_subSections = soup_section.find_all(self.config['subsections']['name'], self.config['subsections']['attrs'])
        # all_paragraphs = soup_section.find_all(self.config['paragraphs']['name'])
        # all_tables = soup_section.find_all(self.config['table-container']['name'], self.config['table-container']['attrs'])
        unwanted_paragraphs = []
        [
            unwanted_paragraphs.extend(capt.find_all("p", recursive=True))
            for capt in all_tables
        ]
        # all_figures = soup_section.find_all(self.config["figure"]["name"], self.config['figure']['attrs'])
        [
            unwanted_paragraphs.extend(capt.find_all("p", recursive=True))
            for capt in all_figures
        ]
        all_paragraphs = [
            para for para in all_paragraphs if para not in unwanted_paragraphs
        ]
        # if self.config['paragraphs']['regex'] is not None:
        # 	for para in all_paragraphs:
        # 		success=True
        # 		for rePattern in self.config['paragraphs']['regex']:
        # 			if para.has_attr(rePattern['attrs']):
        # 				if not re.match(rePattern['regex'], para.get(rePattern['attrs'])):
        # 					success=False
        # 			else:
        # 				success=False
        # 		if success:
        # 			filtered_paragraphs.append(para)
        children = soup_section.findChildren(recursive=False)
        for child in children:
            self.__navigate_children(child, all_subSections, all_paragraphs)

    def __get_references(self, soup_section):
        """
        do somet with references here
        :return:
        """
        all_references = handle_not_tables(self.config["references"], soup_section)
        for ref in all_references:
            self.paragraphs.append(
                references(ref, self.config, self.section_heading).to_dict()
            )

    def __init__(self, config, sectionDict):
        self.config = config
        self.section_heading = (
            sectionDict["headers"][0]
            if "headers" in sectionDict and not sectionDict["headers"] == ""
            else ""
        )
        self.__set_IAO()
        self.subheader = ""
        self.paragraphs = []
        if self.section_heading == "Abbreviations":
            self.__get_abbreviations(sectionDict["node"])
        elif {
            "iao_name": "references section",
            "iao_id": "IAO:0000320",
        } in self.section_type:
            self.__get_references(sectionDict["node"])
        else:
            self.__get_section(sectionDict["node"])

    def to_dict(self):
        return self.paragraphs if self.paragraphs else []



#!/usr/bin/env python3

from datetime import datetime
from operator import itemgetter
from pathlib import Path

import cv2
import pytesseract


class table_image:
    def img2text(self, img, x, y, w, h):
        """
        Function: translate image into texts
        Input: original image, and location of text boxes
        Output: extracted texts
        """

        ROI = img[y - 3 : (y + h + 6), x - 3 : (x + w + 6)]
        # pytesseract.pytesseract.tesseract_cmd = 'D:/Tesseract/tesseract.exe'
        # change the 'lang' here for different traineddata
        text = pytesseract.image_to_string(
            ROI, lang=self.trainedData, config="--psm 6 --oem 3"
        ).strip()
        new_text = text.replace("\n", " ")
        return new_text

    def rm_lines(self, img):
        """
        Function: remove all the horizontal and vertical lines in image and binary it
        Input: original image
        Output: image after preprocessing
        """

        gray = cv2.cvtColor(img, cv2.COLOR_BGR2GRAY)
        binary = cv2.adaptiveThreshold(
            ~gray, 255, cv2.ADAPTIVE_THRESH_GAUSSIAN_C, cv2.THRESH_BINARY, 35, -5
        )
        # binary = cv2.adaptiveThreshold(gray, 255, cv2.ADAPTIVE_THRESH_GAUSSIAN_C, cv2.THRESH_BINARY, 15, -5)
        rows, cols = binary.shape

        # detect horizontal lines
        scale = 40
        kernel = cv2.getStructuringElement(cv2.MORPH_RECT, (cols // scale, 1))
        eroded = cv2.erode(binary, kernel, iterations=1)
        dilatedcol = cv2.dilate(eroded, kernel, iterations=2)

        # detect vertical lines
        scale = 20
        kernel = cv2.getStructuringElement(cv2.MORPH_RECT, (1, rows // scale))
        eroded = cv2.erode(binary, kernel, iterations=1)
        dilatedrow = cv2.dilate(eroded, kernel, iterations=2)

        # merge two groups of lines
        merge = cv2.add(dilatedcol, dilatedrow)
        # comment the next line to save image with detected lines
        # cv2.imwrite("lines.jpg", merge)

        after = cv2.add(gray, merge)
        # comment the next line to save borderless table images
        # cv2.imwrite("borderless.jpg", after)

        return after

    def find_cells(self, img):
        """
        Function: find cells in table images and sort them from top-left to bottom-right
        Input: original image
        Output: ordered table cells, and processed image
        """

        added = cv2.copyMakeBorder(
            img, 10, 10, 10, 10, cv2.BORDER_CONSTANT, value=[255, 255, 255]
        )
        size = added.shape
        # print('added.shape: ', size)
        size[0] * size[1]

        # gray = cv2.cvtColor(added, cv2.COLOR_BGR2GRAY)
        gray = self.rm_lines(img)
        ret, thresh = cv2.threshold(gray, 190, 255, cv2.THRESH_BINARY)
        # thresh2 = cv2.adaptiveThreshold(gray,255,cv2.ADAPTIVE_THRESH_GAUSSIAN_C, cv2.THRESH_BINARY,10,0)
        # comment the next line to save binary tables
        # cv2.imwrite("thresh.jpg", thresh)

        rows, cols = thresh.shape
        scale = 150  # the larger, the rectangles smaller
        # the second parameter of kernel and morphology iterations /
        # need to be fine-tuned according to the image size
        kernel = cv2.getStructuringElement(
            cv2.MORPH_RECT, (cols // scale, rows // scale + 2)
        )
        # Another method for erosion
        # eroded = cv2.morphologyEx(thresh, cv2.MORPH_GRADIENT, kernel, iterations=3)
        # eroded = cv2.bitwise_not(eroded)
        eroded = cv2.erode(thresh, kernel, iterations=3)
        # comment the next line to save images after morphology processing
        # cv2.imwrite(target_dir + '/' + "{}_eroded.jpg".format(pmc), eroded)

        # Add white borders before finding contours
        # eroded = eroded[10:(rows - 10), 10:(cols - 10)]
        eroded = cv2.copyMakeBorder(
            eroded, 10, 10, 10, 10, cv2.BORDER_CONSTANT, value=[255, 255, 255]
        )
        thresh = cv2.copyMakeBorder(
            thresh, 10, 10, 10, 10, cv2.BORDER_CONSTANT, value=[255, 255, 255]
        )
        contours, hierarchy = cv2.findContours(
            eroded, cv2.RETR_TREE, cv2.CHAIN_APPROX_SIMPLE
        )

        # 'cells' save the location and sort
        cells = []
        for c in contours:
            x, y, w, h = cv2.boundingRect(c)
            # case 1：eliminate rectangles that are too thin (might be lines)
            if w > h * 20 or h > w * 20:
                continue
            # case 2：remove a box similar to the whole image
            if (w > size[1] * 0.8) and (h > size[0] * 0.8):
                continue

            # case 3: eliminate small boxes that could be noises
            area = cv2.contourArea(c)
            # method 1: constant area. Does not work on images that are too large or too small
            if area < 250:
                continue
            # method 2: proportional area
            # if area > 0:
            #     if imgarea / area > 20000:
            #         continue

            cells.append((x, y, w, h))

        # To avoid location errors in one line
        cells = sorted(cells, key=itemgetter(1, 0))

        return cells, added, thresh

    def cell2table(self, cells, added, thresh, target_dir, pmc):
        """
        Function: save table texts in several rows
        Input: ordered table cells, and processed image
        Output: table text saved line by line
        """

        # after sort, read cells line by line
        color = (0, 255, 0)  # box color
        table_row = []
        row = []

        for i, (x, y, w, h) in enumerate(cells):
            # print(x, y, w, h)
            cv2.rectangle(added, (x, y), (x + w, y + h), color, 1)
            row.append(cells[i])

            # the last cell, footer or normal cell
            if i == len(cells) - 1:
                # save the last line
                table_row.append(row)
                break

            # newlines: [i] x+w >[i+1] x // [i+1] y > [i] y+h (latter is used, more accurate)
            # minus 5 in case two lines are too close
            if cells[i + 1][1] > cells[i][1] + cells[i][3] - 5:
                table_row.append(row)
                # save a new line
                row = []

        """
        # The block (have not fully developed) is used to recognize section names in the left-most column

        for row in table_row:
            new_row = []
            row.sort(key=lambda x: x[0])

            # start = row[0]
            # # row(i)[0][x]+[w] < row(i+1)[0][x] means the below cell is blank
            # if prev[0] + prev[2] < start[0]:
            # append to the above line

            for (j, (x, y, w, h)) in enumerate(row):
                # print(row)
                # # if cells in the first column have all white pixels below, marked as section name
                # # x<w ensures it is in the first column
                # if j == 0 and x<w and eroded[x, y+2*h] > 0:
                #     # append the first element row[0]
                #     new_row.append(''.join(img2text(thresh, x, y, w, h)))
                #     new_table.append(new_row)
                #     new_row=[]
                #     continue

                new_row.append(img2text(thresh, x, y, w, h))

                # comment to write OCR results directly in the image
                # font = cv2.FONT_HERSHEY_SIMPLEX
                # cv2.putText(added, text, (x, y - 10), font, 1, color, 1);

            table_row.append(new_row)
        """

        for row in table_row:
            row.sort(key=lambda x: x[0])
            for i, (x, y, w, h) in enumerate(row):
                row[i] = self.img2text(thresh, x, y, w, h)

        # cv2.imwrite(target_dir + '/' + "{}_result.jpg".format(pmc), added)

        return table_row

    def text2json(self, table_row):
        """
        Function: save table into a formatted json file
        Input: table text saved line by line
        Output: formatted json file of tables
        """

        identifier = ""
        title = ""
        footer = ""
        superline = ""
        cnt1 = cnt2 = 0  # count to identify the column name line

        for i, row in enumerate(table_row):
            if len(row) == 1:
                if i == 0:
                    while len(table_row[i]) == 1:
                        superline = superline + " " + "".join(table_row[i])
                        i = i + 1
                        cnt1 = cnt1 + 1
                    low = superline.lower()
                    identifier = superline[low.find("table") : low.find("table") + 7]
                    title = superline[low.find("table") + 9 :].strip()
                if i == len(table_row) - 1:
                    superline = ""
                    while len(table_row[i]) == 1:
                        superline = "".join(table_row[i]) + " " + superline
                        i = i - 1
                        cnt2 = cnt2 + 1
                    footer = superline

        # remove titles and footers
        # table_row = table_row[cnt1: len(table_row) - 1 - cnt2]
        table_row = table_row[cnt1 : len(table_row) - cnt2]

        table = {}
        sections = []
        cur_section = {}

        pre_header = []
        pre_superrow = None
        cur_header = ""
        cur_superrow = ""

        for i, row in enumerate(table_row):
            if i == 0:
                cur_header = row
            # elif is_column:
            #     cur_header = row
            elif i != 0 and len(row) == 1:
                if i != len(table_row) - 1:
                    cur_superrow = row

            # # last line start with doi, retain?
            # if ''.join(cells[i]).startswith('doi:'):
            #     break

            # skip blank rows (rarely happen)
            if not any([i for i in row if i not in ["", "None"]]):
                continue

            else:
                if cur_header != pre_header:
                    sections = []
                    pre_superrow = None
                    table = {
                        "identifier": identifier,
                        "title": title,
                        "columns": cur_header,
                        "section": sections,
                        "footer": footer,
                    }
                elif cur_header == pre_header:
                    table["section"] = sections

                if cur_superrow != pre_superrow:
                    cur_section = {"section_name": cur_superrow, "results": []}
                    sections.append(cur_section)
                elif cur_superrow == pre_superrow:
                    cur_section["results"].append(row)

                pre_header = cur_header
                pre_superrow = cur_superrow

        # table_json = {'tables': table}

        return table

    def __reformat_table_json(self, table):
        if table == {}:
            return {}
        offset = 0
        if "title" not in table:
            print("no title")
        tableDict = {
            "inputfile": self.file_name,
            "id": self.tableIdentifier,
            "infons": {},
            "passages": [
                {
                    "offset": 0,
                    "infons": {
                        "section_title_1": "table_title",
                        "iao_name_1": "document title",
                        "iao_id_1": "IAO:0000305",
                    },
                    "text": table["title"],
                }
            ],
        }
        offset += len(table["title"])
        if "caption" in table.keys() and not table["caption"] == "":
            tableDict["passages"].append(
                {
                    "offset": offset,
                    "infons": {
                        "section_title_1": "table_caption",
                        "iao_name_1": "caption",
                        "iao_id_1": "IAO:0000304",
                    },
                    "text": table["caption"],
                }
            )
            offset += len(table["caption"])

        if "section" in table.keys():
            rowID = 2
            rsection = []
            this_offset = offset
            for sect in table["section"]:
                resultsDict = {
                    "table_section_title_1": sect["section_name"],
                    "data_rows": [],
                }
                for resultrow in sect["results"]:
                    colID = 1
                    rrow = []
                    for result in resultrow:
                        resultDict = {
                            "cell_id": f"{self.tableIdentifier}.{rowID}.{colID}",
                            "cell_text": result,
                        }
                        colID += 1
                        offset += len(str(result))
                        rrow.append(resultDict)
                    resultsDict["data_rows"].append(rrow)
                    rowID += 1
                rsection.append(resultsDict)

                columns = []
            for i, column in enumerate(table.get("column_headings", [])):
                columns.append(
                    {
                        "cell_id": f"{self.tableIdentifier}.1.{i + 1}",
                        "cell_text": column,
                    }
                )
            tableDict["passages"].append(
                {
                    "offset": this_offset,
                    "infons": {
                        "section_title_1": "table_content",
                        "iao_name_1": "table",
                        "iao_id_1": "IAO:0000306",
                    },
                    "column_headings": columns,
                    "results_section": rsection,
                }
            )

        if "footer" in table.keys() and not table["footer"] == "":
            tableDict["passages"].append(
                {
                    "offset": offset,
                    "infons": {
                        "section_title_1": "table_footer",
                        "iao_name_1": "caption",
                        "iao_id_1": "IAO:0000304",
                    },
                    "text": table["footer"],
                }
            )
            offset += len(table["footer"])
        return tableDict

    def __init__(self, table_images, base_dir, trainedData="eng"):
        self.trainedData = trainedData
        self.table_raw = []
        self.tables = {
            "source": "Auto-CORPus (tables)",
            "date": f'{datetime.today().strftime("%Y%m%d")}',
            "key": "autocorpus_tables.key",
            "infons": {},
            "documents": [],
        }
        base_dir = Path(base_dir)
        for image_path in table_images:
            image_path = Path(image_path)
            imgname = image_path.name
            self.tableIdentifier = imgname.split("_")[-1].split(".")[0]
            self.file_name = str(image_path.relative_to(base_dir))
            pmc = imgname.stem

            img = cv2.imread(str(image_path))

            cells, added, thresh = self.find_cells(img)
            table_row = self.cell2table(cells, added, thresh, "imagesOut", pmc)
            self.tables["documents"].append(
                self.__reformat_table_json(self.text2json(table_row))
            )

    def to_dict(self):
        return self.tables
